using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Collections;

namespace SevenCitiesOfGoldUtility {
  public partial class Form1 : Form {
    private MapViewer _mapViewer;
    private Sprites _sprites;

    public Form1 () {
      InitializeComponent ();
      _sprites = new Sprites ();
      _mapViewer = new MapViewer ((int) _zoomUI.Value, _fogUI.CheckState, _hintsUI.Checked, _enhancementsUI.Checked, _nativesUI.SelectedIndex, true, _sprites);
      _scoreDgv.DataSource = _mapViewer.ScoresDs;
      _nativesUI.SelectedIndex = 0;
      _fogUI.Checked = false;
      _fogUI.Enabled = false;

      ResizeDrawingSurfaceZoomed ();
      ResizeDrawingSurface ();
    }

    private void ResizeDrawingSurface () {
      //_drawingSurface.AutoScrollMinSize = new Size (_mapViewer.Width, _mapViewer.Height);
      //_hScrollBar.Minimum = 0;
      //_vScrollBar.Minimum = 0;
      //_hScrollBar.Maximum = _mapViewer.Width - _drawingSurface.Width;
      //_vScrollBar.Maximum = _mapViewer.Height - _drawingSurface.Height;
      if ((_drawingSurface.Width - 32) >= _mapViewer.Width) {
        _hScrollBar.Enabled = false;
        _hScrollBar.Value = 0;
      }
      else {
        _hScrollBar.Enabled = true;
        _hScrollBar.LargeChange = _drawingSurface.Width - 32;
        _hScrollBar.Maximum = _mapViewer.Width - 1;
        if (_hScrollBar.Value > (_mapViewer.Width - _drawingSurface.Width + 32)) {
          _hScrollBar.Value = _mapViewer.Width - _drawingSurface.Width + 32;
        }
      }
      if (_drawingSurface.Height >= _mapViewer.Height) {
        _vScrollBar.Enabled = false;
        _vScrollBar.Value = 0;
      }
      else {
        _vScrollBar.Enabled = true;
        _vScrollBar.LargeChange = _drawingSurface.Height;
        _vScrollBar.Maximum = _mapViewer.Height - 1;
        if (_vScrollBar.Value > (_mapViewer.Height - _drawingSurface.Height)) {
          _vScrollBar.Value = _mapViewer.Height - _drawingSurface.Height;
        }
      }
      _drawingSurface.Refresh ();
    }

    private void ResizeDrawingSurfaceZoomed () {
      if (_drawingSurfaceZoomed.Height < _drawingSurfaceZoomed.Width) {
        _mapViewer.ResizeZoomed (_drawingSurfaceZoomed.Height, _drawingSurfaceZoomed.Height);
      }
      else {
        _mapViewer.ResizeZoomed (_drawingSurfaceZoomed.Width, _drawingSurfaceZoomed.Width);
      }

      _hScrollBarZoomed.LargeChange = _mapViewer.WidthZoomed;
      _hScrollBarZoomed.Maximum = 255;
      if (_hScrollBarZoomed.Value > (256 - _mapViewer.WidthZoomed)) {
        _hScrollBarZoomed.Value = 256 - _mapViewer.WidthZoomed;
      }
      _vScrollBarZoomed.LargeChange = _mapViewer.HeightZoomed;
      _vScrollBarZoomed.Maximum = 399;
      if (_vScrollBarZoomed.Value > (400 - _mapViewer.HeightZoomed)) {
        _vScrollBarZoomed.Value = 400 - _mapViewer.HeightZoomed;
      }
      _drawingSurfaceZoomed.Refresh ();
    }

    //private void _drawingSurface_Scroll (object sender, ScrollEventArgs e) {
    //  _drawingSurface.Refresh ();
    //}

    Font _font = new Font ("Arial", 8.0f, FontStyle.Regular, GraphicsUnit.Point);
    private bool _redrawEnabled = true;

    private void drawingSurface_Paint (object sender, PaintEventArgs e) {
      //if (!_redrawEnabled) {
      //  return;
      //}
      Graphics g = e.Graphics;
      //g.DrawImage (_mapViewer.Bitmap, _drawingSurface.AutoScrollPosition.X, _drawingSurface.AutoScrollPosition.Y);
      g.DrawImage (_mapViewer.Bitmap, 16 - _hScrollBar.Value, -_vScrollBar.Value);
      int zoom = (int) _zoomUI.Value;
      g.DrawRectangle (Pens.Yellow, 16 + _hScrollBarZoomed.Value * zoom - _hScrollBar.Value, _vScrollBarZoomed.Value * zoom - _vScrollBar.Value, _mapViewer.WidthZoomed * zoom - 1, _mapViewer.HeightZoomed * zoom - 1);
      g.FillRectangle (Brushes.Black, 0, 0, 16, _drawingSurface.Height);
      g.FillRectangle (Brushes.Black, _drawingSurface.Width - 16, 0, 16, _drawingSurface.Height);
      for (int yBigCell = 0; yBigCell < 25; yBigCell += 2) {
        //int latitude = (int) Math.Floor ((double) ((int) (Math.Abs (yBigCell - 11))) / 2);
        int latitude = (int) Math.Floor ((double) ((int) (Math.Abs (yBigCell - 11))) / 2) * 10;
        //int latitude = ((int) (Math.Floor ((double) ((yBigCell - 11) / 2)))) * 10;
        if (latitude < 0) {
          latitude *= -1;
        }
        int y0 = ((yBigCell + 1) * 16) * zoom - _vScrollBar.Value;
        int y1 = (((yBigCell + 2 + 1) * 16) * zoom) - 1 - _vScrollBar.Value;
        int x0 = 0;
        g.DrawLine (Pens.White, x0, y0, x0 + 15, y0);
        g.DrawLine (Pens.White, x0, y1, x0 + 15, y1);
        g.DrawString (latitude.ToString (), _font, Brushes.White, x0, (y0 + y1) / 2);

        if (_drawingSurface.VerticalScroll.Visible) {
          x0 = _drawingSurface.Width - 32;
        }
        else {
          x0 = _drawingSurface.Width - 16;
        }
        g.DrawLine (Pens.White, x0, y0, x0 + 15, y0);
        g.DrawLine (Pens.White, x0, y1, x0 + 15, y1);
        g.DrawString (latitude.ToString (), _font, Brushes.White, x0, (y0 + y1) / 2);
      }
    }

    private void _drawingSurfaceZoomed_Paint (object sender, PaintEventArgs e) {
      //if (!_redrawEnabled) {
      //  return;
      //}
      Graphics g = e.Graphics;
      g.DrawImage (_mapViewer.BitmapZoomed, (_drawingSurfaceZoomed.Width - _mapViewer.WidthZoomed * Sprites.Width) / 2, (_drawingSurfaceZoomed.Height - _mapViewer.HeightZoomed * Sprites.Height) / 2);
      //g.DrawImage (_mapViewer.BitmapZoomed, 0, 0);
    }

    private void originalMapToolStripMenuItem_Click (object sender, EventArgs e) {
      _mapViewer = new MapViewer ((int) _zoomUI.Value, _fogUI.CheckState, _hintsUI.Checked, _enhancementsUI.Checked, _nativesUI.SelectedIndex, true, _sprites);
      _scoreDgv.DataSource = _mapViewer.ScoresDs;
      _fogUI.Checked = false;
      _fogUI.Enabled = false;
      _redrawEnabled = false;
      _hScrollBarZoomed.Value = 0;
      _vScrollBarZoomed.Value = 0;
      _redrawEnabled = true;
      ResizeDrawingSurface ();
      ResizeDrawingSurfaceZoomed ();
    }

    private void loadMapFromATRToolStripMenuItem_Click (object sender, EventArgs e) {
      OpenFileDialog ofd = new OpenFileDialog ();
      ofd.ReadOnlyChecked = true;
      ofd.DefaultExt = "atr";
      ofd.AddExtension = false;
      ofd.Filter = "ATR files (*.atr)|*.atr";
      //ofd.Title = "";

      DialogResult dg = ofd.ShowDialog ();
      if (dg == DialogResult.OK) {
        _mapViewer = new MapViewer ((int) _zoomUI.Value, _fogUI.CheckState, _hintsUI.Checked, _enhancementsUI.Checked, _nativesUI.SelectedIndex, true, ofd.FileName, _sprites);
        _scoreDgv.DataSource = _mapViewer.ScoresDs;
        _fogUI.Checked = false;
        _fogUI.Enabled = false;
        _redrawEnabled = false;
        _hScrollBarZoomed.Value = 0;
        _vScrollBarZoomed.Value = 0;
        _redrawEnabled = true;
        ResizeDrawingSurface ();
        ResizeDrawingSurfaceZoomed ();
      }
    }

    private void loadSaveGameToolStripMenuItem_Click (object sender, EventArgs e) {
      OpenFileDialog ofd = new OpenFileDialog ();
      ofd.ReadOnlyChecked = true;
      ofd.DefaultExt = "atr";
      ofd.AddExtension = false;
      ofd.Filter = "ATR files (*.atr)|*.atr";
      //ofd.Title = "";

      DialogResult dg = ofd.ShowDialog ();
      if (dg == DialogResult.OK) {
        _mapViewer = new MapViewer ((int) _zoomUI.Value, _fogUI.CheckState, _hintsUI.Checked, _enhancementsUI.Checked, _nativesUI.SelectedIndex, false, ofd.FileName, _sprites);
        _scoreDgv.DataSource = _mapViewer.ScoresDs;
        _fogUI.Checked = true;
        _fogUI.Enabled = true;
        _redrawEnabled = false;
        _hScrollBarZoomed.Value = 0;
        _vScrollBarZoomed.Value = 0;
        _redrawEnabled = true;
        ResizeDrawingSurface ();
        ResizeDrawingSurfaceZoomed ();
      }
    }

    private void exitToolStripMenuItem_Click (object sender, EventArgs e) {
      Close ();
    }

    private void saveMapToPNGToolStripMenuItem_Click (object sender, EventArgs e) {
      SaveFileDialog sfd = new SaveFileDialog ();
      sfd.DefaultExt = "png";
      sfd.AddExtension = true;
      sfd.CheckPathExists = true;
      sfd.Filter = "PNG files (*.png)|*.png";
      sfd.OverwritePrompt = true;
      //sfd.Title = "";

      DialogResult dg = sfd.ShowDialog ();
      if (dg == DialogResult.OK) {
        _mapViewer.SaveToPng (sfd.FileName);
      }
    }

    private void saveFullMapToPNGToolStripMenuItem_Click (object sender, EventArgs e) {
      SaveFileDialog sfd = new SaveFileDialog ();
      sfd.DefaultExt = "png";
      sfd.AddExtension = true;
      sfd.CheckPathExists = true;
      sfd.Filter = "PNG files (*.png)|*.png";
      sfd.OverwritePrompt = true;
      //sfd.Title = "";

      DialogResult dg = sfd.ShowDialog ();
      if (dg == DialogResult.OK) {
        _mapViewer.SaveBigToPng (sfd.FileName);
      }
    }

    private void _drawingSurfaceZoomed_Resize (object sender, EventArgs e) {
      ResizeDrawingSurfaceZoomed ();
      _drawingSurface.Refresh ();
    }

    private void _drawingSurface_MouseDown (object sender, MouseEventArgs e) {
      if (e.Button == MouseButtons.Right) {
        _xMouseOrigin = e.X;
        _yMouseOrigin = e.Y;
        _xOrigin = _hScrollBar.Value;
        _yOrigin = _vScrollBar.Value;
      }
      else {
        DrawingSurfaceClickAndMove (e.X, e.Y, e.Button);
      }
    }

    private void _drawingSurface_MouseMove (object sender, MouseEventArgs e) {
      if (e.Button == MouseButtons.Right) {
        //int zoom = (int) _zoomUI.Value;
        int xMouse = e.X;
        int yMouse = e.Y;
        int xDelta = (_xMouseOrigin - xMouse);
        int yDelta = (_yMouseOrigin - yMouse);

        int xNew = _xOrigin + xDelta;
        if (xNew > (_mapViewer.Width - _drawingSurface.Width + 32)) {
          xNew = _mapViewer.Width - _drawingSurface.Width + 32;
        }
        if (xNew < 0) {
          xNew = 0;
        }
        int yNew = _yOrigin + yDelta;
        if (yNew > (_mapViewer.Height - _drawingSurface.Height)) {
          yNew = _mapViewer.Height - _drawingSurface.Height;
        }
        if (yNew < 0) {
          yNew = 0;
        }

        _redrawEnabled = false;
        _hScrollBar.Value = xNew;
        _vScrollBar.Value = yNew;
        _redrawEnabled = true;
        _drawingSurface.Refresh ();
        //_drawingSurfaceZoomed.Refresh ();
      }
      else {
        DrawingSurfaceClickAndMove (e.X, e.Y, e.Button);
      }
    }

    private void DrawingSurfaceClickAndMove (int xMouse, int yMouse, MouseButtons mouseButtons) {
      if (mouseButtons == MouseButtons.Left) {
        int zoom = (int) _zoomUI.Value;
        int x = (xMouse + _hScrollBar.Value - 16) / zoom - _mapViewer.WidthZoomed / 2;
        int y = (yMouse + _vScrollBar.Value) / zoom - _mapViewer.HeightZoomed / 2;


        //_infoUI.Text = String.Format ("X: {1}{0}Y: {2}{0}xBigCell: {3}{0}yBigCell: {4}{0}xInBigCell: {5}{0}yInBigCell: {6}{0}", System.Environment.NewLine, x, y, x / 16, y / 16, x % 16, y % 16);

        if (x > (256 - _mapViewer.WidthZoomed)) {
          x = 256 - _mapViewer.WidthZoomed;
        }
        if (x < 0) {
          x = 0;
        }
        if (y > (400 - _mapViewer.HeightZoomed)) {
          y = 400 - _mapViewer.HeightZoomed;
        }
        if (y < 0) {
          y = 0;
        }

        _redrawEnabled = false;
        _hScrollBarZoomed.Value = x;
        _vScrollBarZoomed.Value = y;
        _redrawEnabled = true;
        //_mapViewer.MoveClickZoomed (x, y);
        _drawingSurface.Refresh ();
        _drawingSurfaceZoomed.Refresh ();
      }
      else {
        _infoUI.Text = _mapViewer.GetDescription (xMouse + _hScrollBar.Value - 16, yMouse + _vScrollBar.Value);
        //_infoUI.Text = String.Format ("X: {1}{0}Y: {2}{0}xBigCell: {3}{0}yBigCell: {4}{0}xInBigCell: {5}{0}yInBigCell: {6}{0}", System.Environment.NewLine, x, y, x / 16, y / 16, x % 16, y % 16);
      }
    }

    private int _xMouseOrigin;
    private int _yMouseOrigin;
    private int _xOrigin;
    private int _yOrigin;

    private void _drawingSurfaceZoomed_MouseDown (object sender, MouseEventArgs e) {
      if (e.Button == MouseButtons.Right) {
        //_xMouseOrigin = e.X - (_drawingSurfaceZoomed.Width - _mapViewer.WidthZoomed * Sprites.Width) / 2;
        //_yMouseOrigin = e.Y - (_drawingSurfaceZoomed.Height - _mapViewer.HeightZoomed * Sprites.Height) / 2;
        _xMouseOrigin = e.X;
        _yMouseOrigin = e.Y;
        _xOrigin = _hScrollBarZoomed.Value;
        _yOrigin = _vScrollBarZoomed.Value;
      }
    }

    private void _drawingSurfaceZoomed_MouseMove (object sender, MouseEventArgs e) {
      //int xMouse = e.X - (_drawingSurfaceZoomed.Width - _mapViewer.WidthZoomed * Sprites.Width) / 2;
      //int yMouse = e.Y - (_drawingSurfaceZoomed.Height - _mapViewer.HeightZoomed * Sprites.Height) / 2;
      if (e.Button == MouseButtons.Right) {
        int xMouse = e.X;
        int yMouse = e.Y;
        int xDeltaZoomed = (_xMouseOrigin - xMouse) / Sprites.Width;
        int yDeltaZoomed = (_yMouseOrigin - yMouse) / Sprites.Height;

        int xZoomedNew = _xOrigin + xDeltaZoomed;
        if (xZoomedNew < 0) {
          xZoomedNew = 0;
        }
        if (xZoomedNew > (256 - _mapViewer.WidthZoomed)) {
          xZoomedNew = 256 - _mapViewer.WidthZoomed;
        }
        int yZoomedNew = _yOrigin + yDeltaZoomed;
        if (yZoomedNew < 0) {
          yZoomedNew = 0;
        }
        if (yZoomedNew > (400 - _mapViewer.HeightZoomed)) {
          yZoomedNew = 400 - _mapViewer.HeightZoomed;
        }

        _redrawEnabled = false;
        _hScrollBarZoomed.Value = xZoomedNew;
        _vScrollBarZoomed.Value = yZoomedNew;
        _redrawEnabled = true;
        _drawingSurface.Refresh ();
        _drawingSurfaceZoomed.Refresh ();
      }
      else {
        int xMouse = e.X - (_drawingSurfaceZoomed.Width - _mapViewer.WidthZoomed * Sprites.Width) / 2;
        int yMouse = e.Y - (_drawingSurfaceZoomed.Height - _mapViewer.HeightZoomed * Sprites.Height) / 2;
        _infoUI.Text = _mapViewer.GetDescriptionZoomed (xMouse, yMouse);
      }
    }

    private void _drawingSurfaceZoomed_MouseUp (object sender, MouseEventArgs e) {
      if (e.Button == MouseButtons.Right) {

      }
    }

    private void _zoomUI_ValueChanged (object sender, EventArgs e) {
      _mapViewer.Zoom = (int) _zoomUI.Value;
      ResizeDrawingSurface ();
      ResizeDrawingSurfaceZoomed ();
      _drawingSurface.Refresh ();
    }

    private void _fogUI_CheckStateChanged (object sender, EventArgs e) {
      _mapViewer.Fog = _fogUI.CheckState;
      _drawingSurface.Refresh ();
      _drawingSurfaceZoomed.Refresh ();
    }

    private void _hintsUI_CheckedChanged (object sender, EventArgs e) {
      _mapViewer.Hints = _hintsUI.Checked;
      _drawingSurface.Refresh ();
      _drawingSurfaceZoomed.Refresh ();
    }

    private void _enhanceUI_CheckedChanged (object sender, EventArgs e) {
      _mapViewer.Enhancements = _enhancementsUI.Checked;
      _drawingSurface.Refresh ();
      _drawingSurfaceZoomed.Refresh ();
    }

    private void Form1_Load (object sender, EventArgs e) {
      //_scoreDgv.Rows [0].Selected = true;
      //_scoreDgv.Rows [0].Selected = false;
    }

    private void _drawingSurface_Resize (object sender, EventArgs e) {
      ResizeDrawingSurface ();
    }

    private void _hScrollBar_ValueChanged (object sender, EventArgs e) {
      if (_redrawEnabled) {
        _drawingSurface.Refresh ();
      }
    }

    private void _vScrollBar_ValueChanged (object sender, EventArgs e) {
      if (_redrawEnabled) {
        _drawingSurface.Refresh ();
      }
    }

    private void _hScrollBarZoomed_ValueChanged (object sender, EventArgs e) {
      _mapViewer.MoveZoomed (_hScrollBarZoomed.Value, _vScrollBarZoomed.Value);
      if (_redrawEnabled) {
        _drawingSurface.Refresh ();
        _drawingSurfaceZoomed.Refresh ();
      }
    }

    private void _vScrollBarZoomed_ValueChanged (object sender, EventArgs e) {
      _mapViewer.MoveZoomed (_hScrollBarZoomed.Value, _vScrollBarZoomed.Value);
      if (_redrawEnabled) {
        _drawingSurface.Refresh ();
        _drawingSurfaceZoomed.Refresh ();
      }
    }

    private void _nativesUI_SelectedIndexChanged (object sender, EventArgs e) {
      _mapViewer.SelectedNative = _nativesUI.SelectedIndex;
      _drawingSurface.Refresh ();
    }

    private void _scoreDgv_SelectionChanged (object sender, EventArgs e) {
      //_scoreDgv.Rows [0].Selected = true;
      //_scoreDgv.Rows [0].Selected = false;
    }
  }
}