using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace SevenCitiesOfGoldUtility {
  public class MapViewer {
    private int [,] _plots;
    private bool [,] _knownSmallCells;

    private int _nbRivers;
    private int [] _xRivers;
    private int [] _yRivers;
    private int [] _scoreRivers;
    private int [] _typeRivers;

    private int _nbSpecials;
    private int [] _xSmallCellSpecials;
    private int [] _ySmallCellSpecials;
    private int [] _typeSpecials;

    private int _nbMines;
    private int [] _xMines;
    private int [] _yMines;
    private int [] _typeMines;

    private int _nbCamps;
    private int [] _xCamps;
    private int [] _yCamps;
    private int [] _typeCamps;

    private int _nbNatives;
    private int [] _xNatives;
    private int [] _yNatives;
    private int [] _typeNatives;
    private bool [] _knownNatives;
    private int [] _goldNatives;
    private int [] _foodNatives;
    private int [] _menNatives;

    private Bitmap _buffer;
    byte [] _atrDatas;

    private int _zoom;
    private int _width;
    private int _height;
    private CheckState _fog;
    private bool _hints;
    private bool _enhancements;
    private Sprites _sprites;

    private int _xZoomed;
    private int _yZoomed;
    private int _widthZoomed;
    private int _heightZoomed;
    private Bitmap _bufferZoomed;

    private bool _newWorld;

    private int _nbKnownsMaxScore;
    private int _nbKnownsScore;
    private int _nbRiversMaxScore;
    private int _nbRiversScore;
    private int _nbNativesMaxScore;
    private int _nbNativesScore;
    private int _nbMinesMaxScore;
    private int _nbMinesScore;
    private int _nbSpecialsMaxScore;
    private int _nbSpecialsScore;

    // total 1600 small cells
    private int _nbToDiscover;
    private int _selectedNative;

    public MapViewer (int zoom, CheckState fog, bool hints, bool enhancements, int selectedNative, bool newWorld, Sprites sprites) {
      _sprites = sprites;
      _xZoomed = 0;
      _yZoomed = 0;
      _widthZoomed = 8;
      _heightZoomed = 8;
      _bufferZoomed = new Bitmap (_widthZoomed * Sprites.Width, _heightZoomed * Sprites.Height);

      _newWorld = newWorld;

      _atrDatas = Resources.SevenCitiesOfGoldAtrImage;
      SetZoom (zoom);
      _fog = fog;
      _hints = hints;
      _enhancements = enhancements;
      _selectedNative = selectedNative;
      LoadMap ();
    }

    public MapViewer (int zoom, CheckState fog, bool hints, bool enhancements, int selectedNative, bool newWorld, string fileName, Sprites sprites) {
      _sprites = sprites;
      _xZoomed = 0;
      _yZoomed = 0;
      _widthZoomed = 8;
      _heightZoomed = 8;
      _bufferZoomed = new Bitmap (_widthZoomed * Sprites.Width, _heightZoomed * Sprites.Height);

      _newWorld = newWorld;

      System.IO.FileStream stream = new System.IO.FileStream (fileName, System.IO.FileMode.Open);
      //stream.Seek (0x40BD, System.IO.SeekOrigin.Begin);
      _atrDatas = new byte [92176];
      stream.Read (_atrDatas, 0, 92176);
      //for (int level = 1; level <= 8; level++) {
      //  for (int dungeon = 1; dungeon <= 16; dungeon++) {
      //    _seeds [level - 1, dungeon - 1] = b;
      //    b = stream.ReadByte ();
      //  }
      //}
      stream.Close ();
      //int level2 = (int) _levelUI.Value;
      //int dungeon2 = (int) _dungeonUI.Value;
      //_seedUI.Value = _seeds [level2 - 1, dungeon2 - 1];
      SetZoom (zoom);
      _fog = fog;
      _hints = hints;
      _enhancements = enhancements;
      _selectedNative = selectedNative;
      LoadMap ();
    }

    private bool _dirty;
    private bool _dirtyZoomed;

    private void LoadMap () {
      _plots = new int [256, 400];

      //_riverBrush = new SolidBrush (Color.FromArgb (1, 83, 167));
      //_shallowBrush = new SolidBrush (Color.FromArgb (1, 68, 152));
      //_mediumBrush = new SolidBrush (Color.FromArgb (1, 53, 137));
      //_deepBrush = new SolidBrush (Color.FromArgb (1, 38, 132));

      //_riverBrush = new SolidBrush (Color.FromArgb (1, 103, 187));
      //_shallowBrush = new SolidBrush (Color.FromArgb (1, 83, 167));
      //_mediumBrush = new SolidBrush (Color.FromArgb (1, 63, 147));
      //_deepBrush = new SolidBrush (Color.FromArgb (1, 48, 132));


      Build ();
      _dirty = true;
      _dirtyZoomed = true;
    }

    private void SetZoom (int zoom) {
      _zoom = zoom;
      _width = _zoom * 256;
      _height = _zoom * 400;
      _buffer = new Bitmap (_width, _height);
      _dirty = true;
    }

    private void Redraw () {
      Draw (Graphics.FromImage (_buffer));
    }

    private void RedrawZoomed () {
      DrawZoomed (Graphics.FromImage (_bufferZoomed), _xZoomed, _yZoomed, _widthZoomed, _heightZoomed);
      _dirtyZoomed = false;
    }

    private void Build () {
      // Plots
      _nbKnownsMaxScore = ((_atrDatas [0x95D9] << 8) + _atrDatas [0x95D8]) * 10;
      if (_newWorld) {
        _nbKnownsScore = 0;
      }
      else {
        _nbKnownsScore = ((_atrDatas [0x5172] << 8) + _atrDatas [0x5171]);
      }
      int [] sector;
      for (int yBigCell = 0; yBigCell < 25; yBigCell++) {
        for (int xBigCell = 0; xBigCell < 16; xBigCell++) {
          sector = ReadSector (yBigCell * 16 + xBigCell);
          for (int y = 0; y < 16; y++) {
            for (int x = 0; x < 8; x++) {
              _plots [xBigCell * 16 + x * 2 + 0, yBigCell * 16 + y] = (sector [y * 8 + x]) >> 4;
              _plots [xBigCell * 16 + x * 2 + 1, yBigCell * 16 + y] = (sector [y * 8 + x]) & 0x0F;
            }
          }
        }
      }
      // Fog
      _nbToDiscover = 0;
      _knownSmallCells = new bool [32, 50];
      if (_newWorld) {
        for (int y = 0; y < 50; y++) {
          for (int x = 0; x < 4; x++) {
            int a = _atrDatas [0x9510 + y * 4 + x];
            for (int i = 0; i < 8; i++) {
              int bit = 1 - ((a & 0x80) >> 7);
              _knownSmallCells [x * 8 + i, y] = (bit == 1);
              if (bit == 0) {
                _nbToDiscover++;
              }
              a <<= 1;
            }
          }
        }
      }
      else {
        for (int y = 0; y < 50; y++) {
          for (int x = 0; x < 4; x++) {
            int a = _atrDatas [0x51C8 + y * 4 + x];
            for (int i = 0; i < 8; i++) {
              int bit = (a & 0x80) >> 7;
              _knownSmallCells [x * 8 + i, y] = (bit == 1);
              if (bit == 0) {
                _nbToDiscover++;
              }
              a <<= 1;
            }
          }
        }
      }
      // Rivers
      _nbRiversMaxScore = ((_atrDatas [0x95DB] << 8) + _atrDatas [0x95DA]) * 2;
      if (_newWorld) {
        _nbRiversScore = 0;
      }
      else {
        _nbRiversScore = (_atrDatas [0x5174] << 8) + _atrDatas [0x5173];
      }
      _xRivers = new int [80];
      _yRivers = new int [80];
      _scoreRivers = new int [80];
      _typeRivers = new int [80];
      if (_newWorld) {
        _nbRivers = _atrDatas [0x9090] + 1;
        for (int i = 0; i < _nbRivers; i++) {
          _xRivers [i] = _atrDatas [0x9091 + i];
          _yRivers [i] = _atrDatas [0x90E1 + i] + (_atrDatas [0x9131 + i] << 8);
          int score = _atrDatas [0x9181 + i];
          if (score >= 128) {
            score = 256 - score;
            _typeRivers [i] = 1;
          }
          else {
            _typeRivers [i] = 0;
          }
          _scoreRivers [i] = score;
        }
      }
      else {
        _nbRivers = _atrDatas [0x4A10] + 1;
        for (int i = 0; i < 80; i++) {
          _xRivers [i] = _atrDatas [0x4A11 + i];
          _yRivers [i] = _atrDatas [0x4A61 + i] + (_atrDatas [0x4AB1 + i] << 8);
          int score = _atrDatas [0x4B01 + i];
          if (score >= 128) {
            score = 256 - score;
            _typeRivers [i] = 1;
          }
          else {
            _typeRivers [i] = 0;
          }
          _scoreRivers [i] = score;
        }
      }
      // Specials
      if (_newWorld) {
        _nbSpecialsScore = 0;
      }
      else {
        _nbSpecialsScore = ((_atrDatas [0x517A] << 8) + _atrDatas [0x5179]);
      }
      _xSmallCellSpecials = new int [20];
      _ySmallCellSpecials = new int [20];
      _typeSpecials = new int [20];
      if (_newWorld) {
        _nbSpecials = _atrDatas [0x91D1];
        for (int i = 0; i < 20; i++) {
          _xSmallCellSpecials [i] = _atrDatas [0x91D2 + i];
          _ySmallCellSpecials [i] = _atrDatas [0x91E6 + i];
          _typeSpecials [i] = _atrDatas [0x91FA + i];
          if (_knownSmallCells [_xSmallCellSpecials [i], _ySmallCellSpecials [i]]) {
            _xSmallCellSpecials [i] = -1;
            _ySmallCellSpecials [i] = -1;
            _typeSpecials [i] = -1;
          }
        }
      }
      else {
        _nbSpecials = _atrDatas [0x4B51];
        for (int i = 0; i < 20; i++) {
          _ySmallCellSpecials [i] = _atrDatas [0x4B66 + i];
          _xSmallCellSpecials [i] = _atrDatas [0x4B52 + i];
          _typeSpecials [i] = _atrDatas [0x4B7A + i];
          if (_knownSmallCells [_xSmallCellSpecials [i], _ySmallCellSpecials [i]]) {
            _xSmallCellSpecials [i] = -1;
            _ySmallCellSpecials [i] = -1;
            _typeSpecials [i] = -1;
          }
        }
      }
      _nbSpecialsMaxScore = _nbSpecials * 100;
      // Mines
      _nbMinesMaxScore = _atrDatas [0x95DD] * 250;
      if (_nbMinesMaxScore == 0) {
        _nbMinesMaxScore = 20 * 250;
      }
      if (_newWorld) {
        _nbMinesScore = 0;
      }
      else {
        _nbMinesScore = ((_atrDatas [0x5178] << 8) + _atrDatas [0x5177]);
      }
      _nbMines = 20;
      _xMines = new int [_nbMines];
      _yMines = new int [_nbMines];
      _typeMines = new int [_nbMines];
      if (_newWorld) {
        for (int i = 0; i < _nbMines; i++) {
          _typeMines [i] = 0;
          _xMines [i] = _atrDatas [0x9010 + i] * 16 + (_atrDatas [0x9038 + i] & 0x0F);
          _yMines [i] = _atrDatas [0x9024 + i] * 16 + (_atrDatas [0x9038 + i] >> 4);
        }
      }
      else {
        for (int i = 0; i < _nbMines; i++) {
          _typeMines [i] = _atrDatas [0x47C0 + i];
          _xMines [i] = _atrDatas [0x47F4 + i] * 16 + (_atrDatas [0x485C + i] & 0x0F);
          _yMines [i] = _atrDatas [0x4828 + i] * 16 + (_atrDatas [0x485C + i] >> 4);
        }
      }
      // Camps
      if (_newWorld) {
        _nbCamps = 0;
      }
      else {
        _nbCamps = _atrDatas [0x52E3];
        _xCamps = new int [_nbCamps];
        _yCamps = new int [_nbCamps];
        _typeCamps = new int [_nbCamps];
        for (int i = 0; i < _nbCamps; i++) {
          _typeCamps [i] = _atrDatas [0x47D4 + i];
          _xCamps [i] = _atrDatas [0x4808 + i] * 16 + (_atrDatas [0x4870 + i] & 0x0F);
          _yCamps [i] = _atrDatas [0x483C + i] * 16 + (_atrDatas [0x4870 + i] >> 4);
        }
      }
      // Natives
      _nbNativesMaxScore = _atrDatas [0x95DC] * 50;
      if (_nbNativesMaxScore == 0) {
        _nbNativesMaxScore = 200 * 50;
      }
      if (_newWorld) {
        _nbMinesScore = 0;
      }
      else {
        _nbNativesScore = ((_atrDatas [0x5176] << 8) + _atrDatas [0x5175]);
      }
      _nbNatives = 256;
      _xNatives = new int [_nbNatives];
      _yNatives = new int [_nbNatives];
      _typeNatives = new int [_nbNatives];
      _knownNatives = new bool [_nbNatives];
      _goldNatives = new int [_nbNatives];
      _foodNatives = new int [_nbNatives];
      _menNatives = new int [_nbNatives];
      if (_newWorld) {
        for (int i = 0; i < _nbNatives; i++) {
          _xNatives [i] = _atrDatas [0x9210 + i];
          _yNatives [i] = _atrDatas [0x9310 + i] * 2;
          _typeNatives [i] = _atrDatas [0x9410 + i];
          _goldNatives [i] = -1;
          if (_typeNatives [i] == 10) {
            _foodNatives [i] = 0;
            _menNatives [i] = 0;
          }
          else {
            _foodNatives [i] = 0xF0 * 2 * _populationNatives [_typeNatives [i]];
            _menNatives [i] = 0x07 * _populationNatives [_typeNatives [i]];
          }
        }
      }
      else {
        for (int i = 0; i < _nbNatives; i++) {
          _xNatives [i] = _atrDatas [0x4B90 + i];
          _yNatives [i] = _atrDatas [0x4C90 + i] * 2;
          _typeNatives [i] = _atrDatas [0x4D90 + i] & 0x0F;
          _knownNatives [i] = ((_atrDatas [0x4F90 + i] & 0x80) >> 7) == 1;
          //int goldFactor = _typeNatives [i];
          //if (goldFactor == 9) {
          //  goldFactor = 14;
          //}
          //_goldNatives [i] = ((_atrDatas [0x4F90 + i] & 0x78) >> 2) * (goldFactor + 1) * (goldFactor + 1);
          _goldNatives [i] = (_atrDatas [0x4F90 + i] & 0x78) >> 3;
          if (_typeNatives [i] == 10) {
            _foodNatives [i] = 0;
          }
          else {
            _foodNatives [i] = (_atrDatas [0x4D90 + i]) * 2 * _populationNatives [_typeNatives [i]];
          }
          _menNatives [i] = ((_atrDatas [0x4E90 + i] >> 3) & 0x07) * _populationNatives [_typeNatives [i]];
          //_unk4Natives [i] = _atrDatas [0x4E90 + i] & 0x0F;
          //_unk5Natives [i] = _atrDatas [0x4F90 + i] & 0x0F;
        }
      }
    }

    private static int [] _populationNatives = { 2, 4, 8, 10, 12, 16, 20, 32, 25, 36, 1 };

    private int [] ReadSector (int sector) {
      int [] sectorDatas = new int [128];
      for (int i = 0; i < 128; i++) {
        sectorDatas [i] = _atrDatas [0x9610 + sector * 128 + i];
      }
      return (sectorDatas);
    }

    public void Draw (Graphics g) {
      if (!_dirty) {
        return;
      }
      Brush riverBrush = new SolidBrush (Color.FromArgb (1, 83, 167));
      Brush shallowBrush = new SolidBrush (Color.FromArgb (1, 123, 207));
      Brush mediumBrush = new SolidBrush (Color.FromArgb (1, 103, 187));
      Brush deepBrush = new SolidBrush (Color.FromArgb (1, 83, 167));

      //Brush plainBrush = new SolidBrush (Color.FromArgb (203, 121, 37));
      Brush plainBrush = new SolidBrush (Color.FromArgb (169, 87, 3));
      Brush forestBrush = new SolidBrush (Color.FromArgb (16, 67, 0));
      Brush swampBrush = new SolidBrush (Color.FromArgb (90, 80, 60));
      Brush mountainBrush = new SolidBrush (Color.FromArgb (39, 14, 0));

      g.FillRectangle (deepBrush, 0 * 8 * _zoom, 0 * 8 * _zoom, 256 * _zoom, 400 * _zoom);

      // Plots
      for (int y = 0; y < 400; y++) {
        for (int x = 0; x < 256; x++) {
          if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_knownSmallCells [x / 8, y / 8]))) {
            Brush brush;
            int cell = _plots [x, y];
            // Deep
            if (cell == 0) {
              brush = deepBrush;
            }
            // Medium
            else if (cell == 1) {
              if (_enhancements) {
                brush = mediumBrush;
              }
              else {
                brush = deepBrush;
              }
            }
            // Shallow
            else if (cell == 2) {
              if (_enhancements) {
                brush = shallowBrush;
              }
              else {
                brush = deepBrush;
              }
            }
            // Ship
            else if (cell == 3) {
              //if (_enhancements) {
              brush = Brushes.LightPink;
              //}
              //else {
              //  brush = deepBrush;
              //}
            }
            // River Cross
            else if (cell == 4) {
              brush = riverBrush;
            }
            // River Horizontal
            else if (cell == 5) {
              brush = riverBrush;
            }
            // River Top to Left
            else if (cell == 6) {
              brush = riverBrush;
            }
            // River Bottom to Left
            else if (cell == 7) {
              brush = riverBrush;
            }
            // River Vertical
            else if (cell == 8) {
              brush = riverBrush;
            }
            // River Top to Right
            else if (cell == 9) {
              brush = riverBrush;
            }
            // River Bottom to Right
            else if (cell == 10) {
              brush = riverBrush;
            }
            // Plain
            else if (cell == 11) {
              brush = plainBrush;
            }
            // Forest
            else if (cell == 12) {
              brush = forestBrush;
            }
            // Mountain
            else if (cell == 13) {
              brush = mountainBrush;
            }
            // Swamp
            else if (cell == 14) {
              if (_enhancements) {
                brush = swampBrush;
              }
              else {
                brush = forestBrush;
              }
            }
            // Village
            else if (cell == 15) {
              brush = plainBrush;
              //brush = Brushes.White;
            }
            else {
              brush = Brushes.Purple;
            }
            if (brush != deepBrush) {
              g.FillRectangle (brush, x * _zoom, y * _zoom, _zoom, _zoom);
            }
          }
        }
      }
      // Fog
      if ((_fog == CheckState.Checked) || (_fog == CheckState.Indeterminate)) {
        Brush fogBrush;
        if (_fog == CheckState.Checked) {
          fogBrush = Brushes.Black;
        }
        else {
          fogBrush = new SolidBrush (Color.FromArgb (100, 0, 0, 0));
        }
        for (int y = 0; y < 50; y++) {
          for (int x = 0; x < 32; x++) {
            if (!_knownSmallCells [x, y]) {
              g.FillRectangle (fogBrush, x * 8 * _zoom, y * 8 * _zoom, 8 * _zoom, 8 * _zoom);
            }
          }
        }
      }
      if (_hints) {
        // Specials
        if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) {
          Brush specialBrush = new SolidBrush (Color.FromArgb (100, 255, 0, 0));
          for (int i = 0; i < _nbSpecials; i++) {
            if (_typeSpecials [i] != -1) {
              g.FillRectangle (specialBrush, _xSmallCellSpecials [i] * 8 * _zoom + 1, _ySmallCellSpecials [i] * 8 * _zoom + 1, 8 * _zoom - 2, 8 * _zoom - 2);
              g.DrawRectangle (Pens.Black, _xSmallCellSpecials [i] * 8 * _zoom, _ySmallCellSpecials [i] * 8 * _zoom, _zoom * 8 - 1, _zoom * 8 - 1);
            }
          }
        }
        // Rivers
        if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) {
          Brush river2Brush = new SolidBrush (Color.DeepSkyBlue);
          for (int i = 0; i < _nbRivers; i++) {
            if (_xRivers [i] != 0) {
              g.FillRectangle (river2Brush, _xRivers [i] * _zoom, _yRivers [i] * _zoom, _zoom, _zoom);
              g.DrawRectangle (Pens.Black, _xRivers [i] * _zoom - 1, _yRivers [i] * _zoom - 1, _zoom + 1, _zoom + 1);
            }
          }
        }
        // Natives
        Brush nativeBrush = new SolidBrush (Color.WhiteSmoke);
        for (int i = 0; i < _nbNatives; i++) {
          if ((_xNatives [i] != 0) && (_yNatives [i] != 0)) {
            if ((_newWorld) || ((!_newWorld) && ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_knownNatives [i]))))) {
              Brush brush;
              if (_selectedNative == 0) {
                brush = nativeBrush;
              }
              else if ((_selectedNative == 1) && ((_typeNatives [i] == 0) || (_typeNatives [i] == 1))) {
                brush = nativeBrush;
              }
              else if ((_selectedNative == 2) && ((_typeNatives [i] == 2) || (_typeNatives [i] == 3))) {
                brush = nativeBrush;
              }
              else if ((_selectedNative == 3) && ((_typeNatives [i] == 4) || (_typeNatives [i] == 5))) {
                brush = nativeBrush;
              }
              else if ((_selectedNative == 4) && ((_typeNatives [i] == 6) || (_typeNatives [i] == 7))) {
                brush = nativeBrush;
              }
              else if ((_selectedNative == 5) && (_typeNatives [i] == 8)) {
                brush = nativeBrush;
              }
              else if ((_selectedNative == 6) && (_typeNatives [i] == 9)) {
                brush = nativeBrush;
              }
              else if ((_selectedNative == 7) && (_typeNatives [i] == 10)) {
                brush = nativeBrush;
              }
              else {
                brush = Brushes.Gray;
              }
              g.FillRectangle (brush, _xNatives [i] * _zoom, _yNatives [i] * _zoom, _zoom, _zoom);
            }
            if ((_newWorld) || ((!_newWorld) && (((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) && (!_knownNatives [i])))) {
              //if ((_newWorld) || ((!_newWorld) && (!_fog && !_knownNatives [i]))) {
              g.DrawRectangle (Pens.Black, _xNatives [i] * _zoom - 1, _yNatives [i] * _zoom - 1, _zoom + 1, _zoom + 1);
            }
          }
        }
        // Mines
        Brush minesBrush = new SolidBrush (Color.HotPink);
        for (int i = 0; i < _nbMines; i++) {
          if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_typeMines [i] != 0x80))) {
            if (_yMines [i] != 240) {
              g.FillRectangle (minesBrush, _xMines [i] * _zoom, _yMines [i] * _zoom, _zoom, _zoom);
              g.DrawRectangle (Pens.Black, _xMines [i] * _zoom - 1, _yMines [i] * _zoom - 1, _zoom + 1, _zoom + 1);
            }
          }
        }
        // Camps
        Brush campsBrush = new SolidBrush (Color.MediumSeaGreen);
        for (int i = 0; i < _nbCamps; i++) {
          if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_typeCamps [i] <= 0x80))) {
            //if (_yCamps [i] != 240) {
            g.FillRectangle (campsBrush, _xCamps [i] * _zoom, _yCamps [i] * _zoom, _zoom, _zoom);
            g.DrawRectangle (Pens.Black, _xCamps [i] * _zoom - 1, _yCamps [i] * _zoom - 1, _zoom + 1, _zoom + 1);
          }
        }
      }
      //Font font = new Font ("Arial", 8.0f, FontStyle.Regular, GraphicsUnit.Point);
      //for (int yBigCell = 0; yBigCell < 25; yBigCell += 2) {
      //  //int latitude = (int) Math.Floor ((double) ((int) (Math.Abs (yBigCell - 11))) / 2);
      //  int latitude = (int) Math.Floor ((double) ((int) (Math.Abs (yBigCell - 11))) / 2) * 10;
      //  //int latitude = ((int) (Math.Floor ((double) ((yBigCell - 11) / 2)))) * 10;
      //  if (latitude < 0) {
      //    latitude *= -1;
      //  }
      //  int y0 = ((yBigCell + 1) * 16) * _zoom;
      //  int y1 = (((yBigCell + 2 + 1) * 16) * _zoom) - 1;
      //  g.DrawLine (Pens.White, 0, y0, 15, y0);
      //  g.DrawLine (Pens.White, 0, y1, 15, y1);
      //  g.DrawString (latitude.ToString (), font, Brushes.White, 0, (y0 + y1) / 2);
      //  g.DrawLine (Pens.White, _width - 16, y0, _width - 1, y0);
      //  g.DrawLine (Pens.White, _width - 16, y1, _width - 1, y1);
      //  g.DrawString (latitude.ToString (), font, Brushes.White, _width - 16, (y0 + y1) / 2);
      //}
      _dirty = false;
    }

    private Brush _fogBrush = new SolidBrush (Color.FromArgb (100, 0, 0, 0));
    public void DrawZoomed (Graphics g, int xZoomed, int yZoomed, int widthZoomed, int heightZoomed) {
      for (int y = 0; y < heightZoomed; y++) {
        for (int x = 0; x < widthZoomed; x++) {
          if ((_fog == CheckState.Checked) && (!_knownSmallCells [(xZoomed + x) / 8, (yZoomed + y) / 8])) {
            g.FillRectangle (Brushes.Black, x * Sprites.Width, y * Sprites.Height, Sprites.Width, Sprites.Height);
          }
          else {
            int plot = _plots [xZoomed + x, yZoomed + y];
            if (plot <= 2) {
              if (_enhancements) {
                g.DrawImage (_sprites._normals [plot], x * Sprites.Width, y * Sprites.Height);
              }
              else {
                g.DrawImage (_sprites._normals [0], x * Sprites.Width, y * Sprites.Height);
              }
            }
            else if (plot <= 11) {
              g.DrawImage (_sprites._normals [plot], x * Sprites.Width, y * Sprites.Height);
            }
            else if (plot == 12) {
              int type;
              if (((y + _yZoomed) <= 17) || ((y + _yZoomed) >= 349)) {
                type = 2;
              }
              else if (((y + _yZoomed) <= 45) || ((y + _yZoomed) >= 321)) {
                type = 1;
              }
              else {
                type = 0;
              }
              g.DrawImage (_sprites._forests [(xZoomed + x) % 4 + (((yZoomed + y) % 4) << 2), type], x * Sprites.Width, y * Sprites.Height);
            }
            else if (plot == 13) {
              g.DrawImage (_sprites._mountains [((xZoomed + x) % 4) * 3 + (xZoomed + x) % 2 + ((yZoomed + y) % 2)], x * Sprites.Width, y * Sprites.Height);
            }
            else if (plot == 14) {
              g.DrawImage (_sprites._swamps [(xZoomed + x) % 4 + (((yZoomed + y) % 4) << 2)], x * Sprites.Width, y * Sprites.Height);
            }
            else if (plot == 15) {
              int native = GetNative (xZoomed + x, yZoomed + y);
              int camp2 = GetCamp (_xZoomed + x, yZoomed + y);
              if ((camp2 == -1) || ((camp2 != -1) && (((_typeCamps [camp2] >> 4) & 0x03) != 2))) {
                if ((_newWorld) || ((!_newWorld) && ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_knownNatives [native]))))) {
                  g.DrawImage (_sprites._natives [_typeNatives [native]], x * Sprites.Width, y * Sprites.Height);
                  //g.DrawImage (_sprites._natives [_typeNatives [native]], x * Sprites.Width, y * Sprites.Height);
                }
                else {
                  g.DrawImage (_sprites._normals [11], x * Sprites.Width, y * Sprites.Height);
                }
              }
              else {
                g.DrawImage (_sprites._normals [11], x * Sprites.Width, y * Sprites.Height);
              }
            }
            // Mines
            int mine = GetMine (xZoomed + x, yZoomed + y);
            if (mine != -1) {
              if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_typeMines [mine] != 0x80))) {
                g.DrawImage (_sprites._buildings [0], x * Sprites.Width, y * Sprites.Height);
                if (_hints) {
                  g.DrawRectangle (Pens.HotPink, x * Sprites.Width, y * Sprites.Height, Sprites.Width - 1, Sprites.Height - 1);
                }
              }
            }
            // Camps
            int camp = GetCamp (xZoomed + x, yZoomed + y);
            if (camp != -1) {
              if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_typeCamps [camp] <= 0x80))) {
                int typeCamp = _typeCamps [camp];
                if (_typeCamps [camp] == 3) {
                  typeCamp = 3;
                }
                else {
                  typeCamp = (_typeCamps [camp] >> 4) & 0x03;
                }
                g.DrawImage (_sprites._buildings [typeCamp], x * Sprites.Width, y * Sprites.Height);
              }
            }
          if ((_fog == CheckState.Indeterminate) && (!_knownSmallCells [(xZoomed + x) / 8, (yZoomed + y) / 8])) {
            g.FillRectangle (_fogBrush, x * Sprites.Width, y * Sprites.Height, Sprites.Width, Sprites.Height);
            }
            // Rivers
            //if (_hints) {
            //  if (!_fog) {
            //    int river = GetRiver (xZoomed + x, yZoomed + y);
            //    if (river != -1) {
            //      if (_xRivers [river] != 0) {
            //        //g.DrawImage (_sprites._buildings [0], x * Sprites.Width, y * Sprites.Height);
            //        g.DrawRectangle (Pens.DeepSkyBlue, x * Sprites.Width, y * Sprites.Height, Sprites.Width - 1, Sprites.Height - 1);
            //      }
            //    }
            //  }
            //}
          }
        }
      }
      if (_hints) {
        // Specials
        if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) {
          for (int i = 0; i < _nbSpecials; i++) {
            if (_typeSpecials [i] != -1) {
              //g.FillRectangle (specialBrush, _xSmallCellSpecials [i] * 8 * Sprites.Width, _ySmallCellSpecials [i] * 8 * Sprites.Height, Sprites.Width, Sprites.Height);
              //g.FillRectangle (specialBrush, (_xSmallCellSpecials [i] * 8 - _xZoomed) * Sprites.Width + 1, (_ySmallCellSpecials [i] * 8 - _yZoomed) * Sprites.Height + 1, 8 * Sprites.Width - 2, 8 * Sprites.Height - 2);
              g.DrawRectangle (Pens.Red, (_xSmallCellSpecials [i] * 8 - _xZoomed) * Sprites.Width, (_ySmallCellSpecials [i] * 8 - _yZoomed) * Sprites.Height, 8 * Sprites.Width - 1, 8 * Sprites.Height - 1);
            }
          }
        }
        // Rivers
        //Brush specialBrush = new SolidBrush (Color.FromArgb (100, 255, 0, 0));
        if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) {
          for (int i = 0; i < _nbRivers; i++) {
            if (_xRivers [i] != 0) {
              //g.FillRectangle (specialBrush, _xSmallCellSpecials [i] * 8 * Sprites.Width, _ySmallCellSpecials [i] * 8 * Sprites.Height, Sprites.Width, Sprites.Height);
              //g.FillRectangle (specialBrush, (_xSmallCellSpecials [i] * 8 - _xZoomed) * Sprites.Width + 1, (_ySmallCellSpecials [i] * 8 - _yZoomed) * Sprites.Height + 1, 8 * Sprites.Width - 2, 8 * Sprites.Height - 2);
              g.DrawRectangle (Pens.DeepSkyBlue, (_xRivers [i] - _xZoomed) * Sprites.Width, (_yRivers [i] - _yZoomed) * Sprites.Height, Sprites.Width - 1, Sprites.Height - 1);
            }
          }
        }
        // Natives
        //Brush specialBrush = new SolidBrush (Color.FromArgb (100, 255, 0, 0));
        for (int i = 0; i < _nbNatives; i++) {
          if ((_xNatives [i] != 0) && (_yNatives [i] != 0)) {
            if ((_newWorld) || ((!_newWorld) && ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_knownNatives [i]))))) {
              g.DrawRectangle (Pens.WhiteSmoke, (_xNatives [i] - _xZoomed) * Sprites.Width, (_yNatives [i] - _yZoomed) * Sprites.Height, Sprites.Width - 1, Sprites.Height - 1);
            }
          }
        }
        // Camps
        for (int i = 0; i < _nbCamps; i++) {
          if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_typeCamps [i] <= 0x80))) {
            g.DrawRectangle (Pens.MediumSeaGreen, (_xCamps [i] - _xZoomed) * Sprites.Width, (_yCamps [i] - _yZoomed) * Sprites.Height, Sprites.Width - 1, Sprites.Height - 1);
          }
        }
      }
    }

    public int GetNative (int x, int y) {
      for (int i = 0; i < _nbNatives; i++) {
        if ((_xNatives [i] == x) && (_yNatives [i] == y)) {
          return (i);
        }
      }
      return (-1);
    }

    public int GetMine (int x, int y) {
      for (int i = 0; i < _nbMines; i++) {
        if ((_xMines [i] == x) && (_yMines [i] == y)) {
          return (i);
        }
      }
      return (-1);
    }

    public int GetCamp (int x, int y) {
      for (int i = 0; i < _nbCamps; i++) {
        if ((_xCamps [i] == x) && (_yCamps [i] == y)) {
          return (i);
        }
      }
      return (-1);
    }

    public int GetRiver (int x, int y) {
      for (int i = 0; i < _nbRivers; i++) {
        if ((_xRivers [i] == x) && (_yRivers [i] == y)) {
          return (i);
        }
      }
      return (-1);
    }

    private string GetMineDescription (int x, int y) {
      int mine = GetMine (x, y);
      if (mine != -1) {
        if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_typeMines [mine] != 0x80))) {
          return ("Gold: Always [64; 127] * 10");
        }
      }
      return (null);
    }

    private string GetRiverDescription (int x, int y) {
      if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) {
        int river = GetRiver (x, y);
        if (river != -1) {
          if (_xRivers [river] != 0) {
            int scoreRiver = _scoreRivers [river] / 32;
            if (scoreRiver > 2) {
              scoreRiver = 2;
            }
            string description = String.Format ("River: {1} OF A {2} RIVER{0}Score: {3}", System.Environment.NewLine, _nameTypeRivers [_typeRivers [river]], _nameScoreRivers [scoreRiver], _scoreRivers [river] * 2);
            return (description);
          }
        }
      }
      return (null);
    }

    private string GetNativeDescription (int x, int y) {
      int native = GetNative (x, y);
      if (native != -1) {
        int camp2 = GetCamp (x, y);
        if ((camp2 == -1) || ((camp2 != -1) && (((_typeCamps [camp2] >> 4) & 0x03) != 2))) {
          if ((_newWorld) || ((!_newWorld) && ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate) || ((_fog == CheckState.Checked) && (_knownNatives [native]))))) {
            string gold;
            int goldFactor = _typeNatives [native];
            if (goldFactor == 9) {
              goldFactor = 14;
            }
            if (_goldNatives [native] == -1) {
              if (_typeNatives [native] >= 9) {
                gold = (15 * 2 * (goldFactor + 1) * (goldFactor + 1)).ToString ();
              }
              else {
                gold = String.Format ("Random [{0}, 15] * {1} ~= {2} to {3}", _typeNatives [native], 2 * (goldFactor + 1) * (goldFactor + 1), _typeNatives [native] * 2 * (goldFactor + 1) * (goldFactor + 1), 15 * 2 * (goldFactor + 1) * (goldFactor + 1));
              }
            }
            else {
              gold = (_goldNatives [native] * 2 * (goldFactor + 1) * (goldFactor + 1)).ToString ();
            }
            string description = String.Format ("Native: {1}{0}Men: {2}{0}Food: {3}{0}Gold: {4}{0}Score: {5}", System.Environment.NewLine, _nameNatives [_typeNatives [native]], _menNatives [native], _foodNatives [native], gold, "Always 50");
            return (description);
          }
        }
      }
      return (null );
    }

    private static string [] _nameSpecials = { "DEEP CANYONS" , "BROAD PLAINS" , "RICH PRAIRIES" , "    A GREAT LAKE" , "GIANT FORESTS" , "LUSH JUNGLES" , "VAST MOUNTAINS" , "TOWERING TREES" , "SWAMP-LANDS" };
    private static string [] _nameTypeRivers = { "SOURCE", "MOUTH" };
    private static string [] _nameScoreRivers = { "MINOR", "MEDIUM", "MAJOR" } ;
    private static string [] _nameNatives = { "A HUNTING TRIBE" , "A HUNTING TRIBE" , "A FARMING TRIBE" , "A FARMING TRIBE *" , "A PUEBLO TOWN" , "A PUEBLO TOWN *" , "AN AZTEC CITY" , "AN AZTEC CITY *" , "AN INCA CITY" , "THE INCA CAPITAL" , "A LOST CITY" };
    private string _description = "";

    public string GetDescription (int x, int y) {
      int xPlot = x / _zoom;
      int yPlot = y / _zoom;
      // Mines
      string mineDescription = GetMineDescription (xPlot, yPlot);
      if (mineDescription != null) {
        _description = mineDescription;
        return (_description);
      }

      // Rivers
      string riverDescription = GetRiverDescription (xPlot, yPlot);
      if (riverDescription != null) {
        _description = riverDescription;
        return (_description);
      }

      // Natives
      string nativeDescription = GetNativeDescription (xPlot, yPlot);
      if (nativeDescription != null) {
        _description = nativeDescription;
        return (_description);
      }

      // Specials
      if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) {
        for (int i = 0; i < _nbSpecials; i++) {
          if (_typeSpecials [i] != -1) {
            int x0 = _xSmallCellSpecials [i] * 8 * _zoom;
            int y0 = _ySmallCellSpecials [i] * 8 * _zoom;
            if ((x >= x0) && (x <= (x0 + _zoom * 8 - 1)) &&
                (y >= y0) && (y <= (y0 + _zoom * 8 - 1))) {
              _description = String.Format ("Special: {1}{0}Score: {2}", System.Environment.NewLine, _nameSpecials [_typeSpecials [i]], "Always 100");
              return (_description);
            }
          }
        }
      }
      return (_description);
    }

    public string GetDescriptionZoomed (int x, int y) {
      int xPlot = x / Sprites.Width;
      int yPlot = y / Sprites.Height;
      if ((xPlot < 0) || (yPlot < 0) || (xPlot >= _widthZoomed) || (yPlot >= _heightZoomed)) {
        return (_description);
      }
      xPlot += _xZoomed;
      yPlot += _yZoomed;
      // Mines
      string mineDescription = GetMineDescription (xPlot, yPlot);
      if (mineDescription != null) {
        _description = mineDescription;
        return (_description);
      }

      // Rivers
      string riverDescription = GetRiverDescription (xPlot, yPlot);
      if (riverDescription != null) {
        _description = riverDescription;
        return (_description);
      }

      // Natives
      string nativeDescription = GetNativeDescription (xPlot, yPlot);
      if (nativeDescription != null) {
        _description = nativeDescription;
        return (_description);
      }
      // Specials
      if ((_fog == CheckState.Unchecked) || (_fog == CheckState.Indeterminate)) {
        for (int i = 0; i < _nbSpecials; i++) {
          if (_typeSpecials [i] != -1) {
            int x0 = _xSmallCellSpecials [i] * 8;
            int y0 = _ySmallCellSpecials [i] * 8;
            if ((xPlot >= x0) && (xPlot <= (x0 + 8 - 1)) &&
                (yPlot >= y0) && (yPlot <= (y0 + 8 - 1))) {
              _description = String.Format ("Special: {1}{0}Score: {2}", System.Environment.NewLine, _nameSpecials [_typeSpecials [i]], "Always 100");
              return (_description);
            }
          }
        }
      }
      return (_description);
    }

    public Bitmap Bitmap {
      get {
        if (_dirty) {
          Redraw ();
        }
        return (_buffer);
      }
    }

    public Bitmap BitmapZoomed {
      get {
        if (_dirtyZoomed) {
          RedrawZoomed ();
        }
        return (_bufferZoomed);
      }
    }

    public int Width {
      get {
        return (_width);
      }
    }

    public int Height {
      get {
        return (_height);
      }
    }

    public int WidthZoomed {
      get {
        return (_widthZoomed);
      }
    }

    public int HeightZoomed {
      get {
        return (_heightZoomed);
      }
    }

    public void SaveToPng (string fileName) {
      _buffer.Save (fileName, ImageFormat.Png);
    }

    public void SaveBigToPng (string fileName) {
      Bitmap buffer = new Bitmap (256 * Sprites.Width, 400 * Sprites.Height);
      DrawZoomed (Graphics.FromImage (buffer), 0, 0, 256, 400);
      buffer.Save (fileName, ImageFormat.Png);
      buffer.Dispose ();
    }

    public int Zoom {
      set {
        SetZoom (value);
        _dirty = true;
      }
    }

    public CheckState Fog {
      set {
        _fog = value;
        _dirty = true;
        _dirtyZoomed = true;
      }
    }

    public bool Hints {
      set {
        _hints = value;
        _dirty = true;
        _dirtyZoomed = true;
      }
    }

    public bool Enhancements {
      set {
        _enhancements = value;
        _dirty = true;
        _dirtyZoomed = true;
      }
    }

    public int SelectedNative {
      set {
        _selectedNative = value;
        _dirty = true;
      }
    }

    //public void MoveClickZoomed (int x, int y) {
    //  _xZoomed = x - 16 / _zoom - _widthZoomed / 2;
    //  _yZoomed = y - _heightZoomed / 2;
    //  LimitZoomed ();
    //  _dirtyZoomed = true;
    //}

    public void MoveZoomed (int x, int y) {
      _xZoomed = x;
      _yZoomed = y;
      LimitZoomed ();
      _dirtyZoomed = true;
    }

    public void LimitZoomed () {
      if (_xZoomed > (256 - _widthZoomed)) {
        _xZoomed = 256 - _widthZoomed;
      }
      if (_xZoomed < 0) {
        _xZoomed = 0;
      }
      if (_yZoomed > (400 - _heightZoomed)) {
        _yZoomed = 400 - _heightZoomed;
      }
      if (_yZoomed < 0) {
        _yZoomed = 0;
      }
    }

    public void ResizeZoomed (int width, int height) {
      _widthZoomed = width / Sprites.Width;
      _heightZoomed = height / Sprites.Height;
      LimitZoomed ();
      _bufferZoomed = new Bitmap (_widthZoomed * Sprites.Width, _heightZoomed * Sprites.Height);
      _dirtyZoomed = true;
    }

    public List<ScoreDs> ScoresDs {
      get {
        List<ScoreDs> scoresDs = new List<ScoreDs> ();
        scoresDs.Add (new ScoreDs ("Land", _nbKnownsScore, _nbKnownsMaxScore));
        scoresDs.Add (new ScoreDs ("Rivers", _nbRiversScore, _nbRiversMaxScore));
        scoresDs.Add (new ScoreDs ("Natives", _nbNativesScore, _nbNativesMaxScore));
        scoresDs.Add (new ScoreDs ("Mines", _nbMinesScore, _nbMinesMaxScore));
        scoresDs.Add (new ScoreDs ("Special", _nbSpecialsScore, _nbSpecialsMaxScore));
        return (scoresDs);
      }
    }
  }
}
